/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
// Original: 
// Minimal 8 Bit CPU (VHDL version)
//  rev 15102001
//  t.boescke@tuhh.de
//
// Instruction set
//   NOR : 00AAAAAA : Accu <= ~( Accu | mem[AAAAAA] )
//   ADD : 01AAAAAA : Accu <= Accu + mem[AAAAAA], update carry
//   STA : 10AAAAAA : mem[AAAAAA] <= Accu
//   JCC : 11DDDDDD : PC <= DDDDDD if (carry == 0) , carry <= 0
//
// Macro set
//   CLR      : NOR allone                 : 
//   LDA mem  : NOR allone, ADD mem        : Accu <= memory[mem]
//   NOT      : Nor zero                   : 
//   JMP dst  : JCC dst, JCC dst           : 
//   JCS dst  : JCC *+2 , JCC dst          : Jump if carry set
//   SUB mem  : NOR zero, ADD mem ADD one  : Accu -= memory[mem]
//       memory[allone] : 0xFF
//       memory[zero]   : 0x00
//       memory[one]    : 0x01
//
//  XO2-256      MCPU MODULE     XO2-256
//  27  (SDA)   1            24  VCC  
//  28  (SCL)   2            23  O_PORT_0 25
//  29  (TMS)   3            22  O_PORT_1 23
//  30  (TCK)   4            21  O_PORT_2 21
//  32  (TDI)   5            20  O_PORT_3 20
//   1  (TDO)   6            19  O_STB    17
//   4          7            18           16
//   5 O_PORT_7 8            17  N.C. (JTAGENB)
//   8 O_PORT_6 9            16           14
//   9 O_PORT_5 10           15  RST      13
//  10 O_PORT_4 11           14  CLK      12
//       GND    12           13  CLK_OUT  11

`define USE_EFB
`define HAVE_OPORT

`define RAM_INCLUDE_DAT	// must
//`define RAM_USE_CLK	// selectable
`define ROM_USE_CLK	// must

module mcpu (
`ifdef USE_EFB
  		input CLK_2X
`else
		input CLK
`endif
		,input RST // active high
                ,inout SCL
                ,inout SDA
`ifdef HAVE_OPORT
                ,output [7:0] O_PORT
                ,output O_STB
`endif
	);

    wire [7:0] mem_data;
    wire [7:0] code_data;

`ifdef USE_EFB
    wire [7:0] I_DATA;
`else
    wire [7:0] I_DATA = mem_data[7:0];
`endif

    reg [8:0] akku = 0;
    wire [5:0] adreg;
    reg [5:0] pc = 0;
    reg [2:0] inst;

    wire [7:0] O_DATA = akku[7:0];
    wire WE = (inst == 3'b110) & ~RST;

`ifdef HAVE_OPORT
    wire OPORT_SEL = (adreg[5:2] == 4'b1100) ;

    reg [7:0] r_oport = 0;
    assign O_PORT [7:0] = r_oport[7:0];

    assign O_STB = (WE & OPORT_SEL) & (adreg[1:0] == 2'b11);

  `ifdef RAM_USE_CLK
    always @(posedge CLK)
    begin
	if (RST) r_oport <= 0;
	else if (O_STB) r_oport <= akku[7:0];
    end
  `else
    always @(posedge O_STB)
    begin
	r_oport <= akku[7:0];
    end
  `endif

`else
    wire OPORT_SEL = 1'b0;
`endif
    wire WB_SEL = adreg[5:4] != 2'b11 ;
    wire MEM_SEL = ~WB_SEL & ~OPORT_SEL;

`ifdef USE_EFB
    assign I_DATA[7:0] = ~WB_SEL ? mem_data[7:0] : wb_data_l[7:0];

    reg master_clk = 1'b0;
    wire CLK = master_clk;
    
    reg [1:0] wb_stat = 0;
    wire [7:0] wb_adr = { 2'b01, adreg[5:0] };
    wire [7:0] wb_data;
    reg [7:0] wb_data_l = 0;

    wire wb_we = (code_data[7:6] == 2'b10) & ~RST;
    wire wb_stb = WB_SEL & (code_data[7:6] == 2'b00);
    reg wb_cyc = 0;

    wire RST_NEG = ~RST;
    EFB_tc_i2c efb_impl (.wb_clk_i(CLK_2X), .wb_rst_i(RST)
    		  , .wb_adr_i(wb_adr), .wb_dat_i(O_DATA), .wb_dat_o(wb_data)
		  , .wb_we_i(wb_we), .wb_cyc_i(wb_cyc), .wb_stb_i(wb_stb)
		  , .tc_clki(CLK), .tc_rstn(RST_NEG)
		  , .i2c1_scl(SCL), .i2c1_sda(SDA)
		);
`endif

`ifdef RAM_USE_CLK
    reg r_we = 1'b0;
    always @(posedge CLK)
    begin
	r_we <= (inst == 0) & (code_data[7:6] == 2'b10) & ~RST;
    end
    wire WE_mem = r_we & MEM_SEL;
`else
    wire WE_mem = WE & MEM_SEL;
`endif

    wire [5:0] mem_addr = { 2'b00 , adreg[3:0] };
    sram sram_impl (.CLK(CLK), .WEB(WE_mem), .ADDRB(mem_addr[5:0])
			, .DIB(O_DATA[7:0]) , .DOB(mem_data[7:0]) );

    wire rom_rd = (inst == 0);
    rom rom_impl (.CLK(CLK), .RD(rom_rd), .ADDRA(pc[5:0]), .DOA(code_data[7:0]) );

`ifdef USE_EFB
    always @(posedge CLK_2X)
    begin
	if (wb_stat == 3)
	  begin
		master_clk <= 1'b1;
		wb_stat <= 0;
		wb_cyc <= 1'b0;
	  end
	else if (wb_stat == 0)
	  begin
		master_clk <= 1'b0;
		wb_stat <= WB_SEL ? 1 : 3;
		wb_cyc <= WB_SEL;
	  end
	else if (wb_stat == 1)
	  begin
		master_clk <= 1'b1;
		wb_stat <= 2;
		wb_cyc <= WB_SEL;
		wb_data_l <= wb_data;
	  end
	else
	  begin
		master_clk <= 1'b0;
		wb_stat <= 3;
		wb_cyc <= 1'b0;
	  end
    end
`endif


//               000
//         fetch & load I_DATA
//
//    100     101    110     111  
//    NOR     ADD    STA     JCC    
// 
//               000
//         fetch & load I_DATA
//
//
    assign adreg[5:0] = code_data[5:0];
    always @(posedge CLK)
    begin
	if (RST)
	    begin
		inst <= 0;
		akku <= 0;
		pc <= 0;
	    end
	else
	    begin
		if (inst == 0) // fetch & load
		    begin
			inst <= { 1'b1 , code_data[7:6] };
		    end
		else	inst <= 0;

		//   JCC : 11DDDDDD : PC <= DDDDDD if (carry == 0) , carry <= 0
		if ((inst == 3'b111 ) & ~akku[8])
			pc <= adreg;
	        else if (inst != 0)
			pc <= pc + 1;

		//   NOR : 00AAAAAA : Accu <= ~( Accu | mem[AAAAAA] )
		if (inst == 3'b100 )
			akku[7:0] <= ~( akku[7:0] | I_DATA );
		//   ADD : 01AAAAAA : Accu <= Accu + mem[AAAAAA], update carry
		else if (inst == 3'b101 )
			akku <= { 1'b0 , akku[7:0] } + { 1'b0, I_DATA };
		else if  (inst == 3'b111)
			akku[8] <= 1'b0;
	    end
    end

endmodule

/*               _______         _______         _______
 * CLK    ______|       |_______|       |_______|       |______|
 * 
 * code_data   01AAAAAA |          01BBBBBB             |
 * inst    000  |      101      |     000       |     110
 *                                               ______________
 * WE    _______________________________________|              |__
 * 
 *                                              @
 * (RAM_USE_CLK)
 *                                       ______________
 * WE_mem ______________________________|              |__
 *              @               @               @               @
 * 
*/
module sram # (
                parameter SIZE = 16
//              parameter SIZE = 64
//              parameter SIZE = 128
//		parameter SIZE = 256
) (
                input CLK,
		input WEB,
                input [5:0] ADDRB,
                input [7:0] DIB,
                output [7:0] DOB
);
    reg  [7:0] mem [0:SIZE-1];
`ifdef RAM_USE_CLK
    reg [5:0] r_addr;
    always @(posedge CLK)
    if (SIZE == 16)
      begin
        if (WEB) mem[ADDRB[3:0]] <= DIB;
	r_addr <= { 2'b00, ADDRB[3:0] };
      end
    else
      begin
        if (WEB) mem[ADDRB] <= DIB;
	r_addr <= ADDRB;
      end
    assign DOB = mem[r_addr];
`else
    always @(posedge WEB)
    begin
        mem[ADDRB] <= DIB;
    end
    assign DOB = mem[ADDRB];
`endif

`ifdef RAM_INCLUDE_DAT
  initial
    begin
	if (SIZE == 64)
           $readmemh("mcpu_ram64.mem" , mem, 0 , SIZE-1);
	else
           $readmemh("mcpu_ram.mem" , mem, 0 , SIZE-1);
    end
`endif

endmodule

module rom # (
                parameter SIZE = 64
) (
                input CLK,
                input RD,
                input [5:0] ADDRA,
                output [7:0] DOA
);
    reg  [11:0] mem [0:SIZE-1];
`ifdef ROM_USE_CLK
    reg [5:0] r_addr;
    always @(negedge CLK)
    begin
	if (RD) r_addr <= ADDRA;
    end
    assign DOA = mem[r_addr];
`else
    assign DOA = mem[ADDRA];
`endif

  initial
    begin
	if (SIZE == 32)
	    $readmemh("mcpu_code32.mem" , mem, 0 , SIZE-1);
	else
	    $readmemh("mcpu_code.mem" , mem, 0 , SIZE-1);
    end

endmodule
