/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
// Original: 
// Minimal 8 Bit CPU (VHDL version)
//  rev 15102001
//  t.boescke@tuhh.de
//
// Instruction set
//   00SSSAAAAAAA : 
//   00000 : ADD (wo carry)
//   00010 : ADC (with carry)
//   00100 : SUB (wo carry)
//   00110 : SBC (with carry)
//   00001 : AND
//   00011 : XOR
//   00101 : OR
//   00111 : LD
//
//   01SSDDDDDDDD : immidiate 
//   0100 : ANDI
//   0101 : XORI
//   0110 : ORI
//   0111 : LDI
//
//   1000xAAAAAAA : ST
//
//   1001xAAAAAAA : ASR
//   1010xAAAAAAA : ROL
//   1011xAAAAAAA : ROR
//
//   11xxCAAAAAAA : J after load carry
//   1100CAAAAAAA : JCS
//   1101CAAAAAAA : JCC
//   1110CAAAAAAA : JEQ
//   1111CAAAAAAA : JNE
//
//  macros:
//   SEC : set carry   JCS *+1 (and C=1)
//   CLC : clear carry JCC *+1 (and C=0)
//   JMP adr :         JCC adr (and C=0), JCC adr (and C=0)
//
//  XO2-256      MCPU MODULE     XO2-256
//  27  (SDA)   1            24  VCC  
//  28  (SCL)   2            23  O_PORT_0 25
//  29  (TMS)   3            22  O_PORT_1 23
//  30  (TCK)   4            21  O_PORT_2 21
//  32  (TDI)   5            20  O_PORT_3 20
//   1  (TDO)   6            19  O_STB    17
//   4          7            18           16
//   5 O_PORT_7 8            17  N.C. (JTAGENB)
//   8 O_PORT_6 9            16           14
//   9 O_PORT_5 10           15  RST      13
//  10 O_PORT_4 11           14  CLK      12
//       GND    12           13  CLK_OUT  11

`define USE_EFB
`define HAVE_OPORT

`define RAM_INCLUDE_DAT	// must
//`define RAM_USE_CLK	// selectable
//`define ROM_USE_CLK	// must

module mcpu (
`ifdef USE_EFB
		input CLK_2X
`else
		input CLK
`endif
		,input RST // active high
                ,inout SCL
                ,inout SDA
`ifdef HAVE_OPORT
                ,output [7:0] O_PORT
                ,output O_STB
`endif
	);

    wire [7:0] mem_data;
    wire [11:0] code_data;

`ifdef USE_EFB
    wire [7:0] I_DATA;
`else
    wire [7:0] I_DATA = mem_data[7:0];
`endif
    reg [1:0] flags = 0;
    wire C = flags[0];
    wire Z = flags[1];

    reg [7:0] akku = 0;
    wire [6:0] adreg;
    reg [6:0] pc = 0;
    reg [2:0] inst;

    wire [7:0] O_DATA = akku[7:0];
    wire WE = (inst == 3'b110) & ~RST;

`ifdef HAVE_OPORT
    wire OPORT_SEL = (adreg[6:2] == 5'b01100) ;

    reg [7:0] r_oport;
    //assign O_PORT [7] = r_oport[7];
    //assign O_PORT [5:0] = r_oport[5:0];
    assign O_PORT [7:0] = r_oport[7:0];

    assign O_STB = (WE & OPORT_SEL) & (adreg[1:0] == 2'b11);

  `ifdef RAM_USE_CLK
    always @(posedge CLK)
    begin
	if (RST) r_oport <= 0;
	else if (O_STB) r_oport <= akku[7:0];
    end
  `else
    always @(posedge O_STB)
    begin
	r_oport <= akku[7:0];
    end
  `endif

`else
    wire OPORT_SEL = 1'b0;
`endif
    wire WB_SEL = (adreg[6] == 1'b0) & (adreg[5:4] != 2'b11) ;
    wire MEM_SEL = ~WB_SEL & ~OPORT_SEL;

`ifdef USE_EFB
    assign I_DATA[7:0] = ~WB_SEL ? mem_data[7:0] : wb_data_l[7:0];

    reg master_clk = 1'b0;
    wire CLK = master_clk;
    
    reg [1:0] wb_stat = 0;
    wire [7:0] wb_adr = { 2'b01, adreg[5:0] };
    wire [7:0] wb_data;
    reg [7:0] wb_data_l;

    wire wb_we = (code_data[11:10] == 2'b10) & ~RST;
    wire wb_stb = WB_SEL & (code_data[11:10] == 2'b00);
    reg wb_cyc = 0;

    wire RST_NEG = ~RST;
    EFB_tc_i2c efb_impl (.wb_clk_i(CLK_2X), .wb_rst_i(RST)
    		  , .wb_adr_i(wb_adr[7:0]), .wb_dat_i(O_DATA[7:0]), .wb_dat_o(wb_data[7:0])
		  , .wb_we_i(wb_we), .wb_cyc_i(wb_cyc), .wb_stb_i(wb_stb)
		  , .tc_clki(CLK), .tc_rstn(RST_NEG)
		  , .i2c1_scl(SCL), .i2c1_sda(SDA)
		);
`endif

`ifdef RAM_USE_CLK
    reg r_we = 1'b0;
    always @(posedge CLK)
    begin
	r_we <= (inst == 0) & (code_data[11:10] == 2'b10) & ~RST;
    end
    wire WE_mem = r_we & MEM_SEL;
`else
    wire WE_mem = WE & MEM_SEL;
`endif
    wire [1:0] mem_addr_hi = adreg[5:4] + 1;
    wire [5:0] mem_addr = { mem_addr_hi[1:0] , adreg[3:0] };
    sram sram_impl (.CLK(CLK), .WEB(WE_mem), .ADDRB(mem_addr[5:0])
			, .DIB(O_DATA[7:0]) , .DOB(mem_data[7:0]) );

    wire rom_rd = (inst == 0);
    rom rom_impl (.CLK(CLK), .RD(rom_rd), .ADDRA(pc[6:0]), .DOA(code_data[11:0]) );

    wire [7:0] F;
    wire [2:0] S = { code_data[7], code_data[9:8]};
    wire [1:0] ES = 2'b00;

    rtavr_alu alu_impl (.A(akku[7:0]), .B(I_DATA[7:0]), .F(F[7:0])
		, .C_in(C), .Z_in(1'b1), .C_out(C_out), .Z_out(Z_out)
		, .S(S[2:0]), .ES(ES[1:0]) );

    wire [7:0] IMM;
    wire [2:0] S_I = { 1'b1, code_data[9:8]};
    rtavr_alu alu_impl_I (.A(akku[7:0]), .B(code_data[7:0]), .F(IMM[7:0])
		, .C_in(1'b0), .Z_in(1'b1)
		, .S(S_I[2:0]), .ES(ES[1:0]) );

`ifdef USE_EFB
    always @(posedge CLK_2X)
    begin
	if (wb_stat == 3)
	  begin
		master_clk <= 1'b1;
		wb_stat <= 0;
		wb_cyc <= 1'b0;
	  end
	else if (wb_stat == 0)
	  begin
		master_clk <= 1'b0;
		wb_stat <= WB_SEL ? 1 : 3;
		wb_cyc <= WB_SEL;
	  end
	else if (wb_stat == 1)
	  begin
		master_clk <= 1'b1;
		wb_stat <= 2;
		wb_cyc <= WB_SEL;
		wb_data_l <= wb_data;
	  end
	else
	  begin
		master_clk <= 1'b0;
		wb_stat <= 3;
		wb_cyc <= 1'b0;
	  end
    end
`endif

    assign adreg[6:0] = code_data[6:0];
    always @(posedge CLK)
    begin
	if (RST)
	    begin
		inst <= 0;
		akku <= 0;
		pc <= 0;
		flags <= 0;
	    end
	else
	    begin
		if (inst == 0) // fetch & load
		    begin
			if ( (code_data[11:10] == 2'b10) 
			   & (code_data[9:8] != 2'b00) )
				inst <= { 1'b0 , code_data[9:8] };
			else	inst <= { 1'b1 , code_data[11:10] };
		    end
		else	inst <= 0;

		if ((inst == 3'b111 ) & 
		    ( (~code_data[9] & (C ^ code_data[8]) )
		     | (code_data[9] & (Z ^ code_data[8]) ) ) )
			pc <= adreg;
	        else if (inst != 0)
			pc <= pc + 1;

		if (inst == 3'b100)
		    begin
			akku[7:0] <= F[7:0];
			if (~S[2]) flags[0] <= C_out;
			flags[1] <= Z_out;
		    end
		if (inst == 3'b101)
		    begin
			akku[7:0] <= IMM[7:0];
			flags[1] <= Z_out;
		    end
		else if  (inst == 3'b111)
		    begin
			flags[0] <= code_data[7];
		    end
		else if  (inst == 3'b001) // ASR
		    begin
			flags[0] <= akku[0];
			akku[7:0] <= { akku[7], akku[7:1] };
		    end
		else if  (inst == 3'b010) // ROL
		    begin
			flags[0] <= akku[7];
			akku[7:0] <= { akku[6:0] , C } ;
		    end
		else if  (inst == 3'b011) // ROR
		    begin
			flags[0] <= akku[0];
			akku[7:0] <= { C , akku[7:1] };
		    end
	    end
    end

endmodule

/*               _______         _______         _______
 * CLK    ______|       |_______|       |_______|       |______|
 * 
 * code_data   01AAAAAA |          01BBBBBB             |
 * inst    000  |      101      |     000       |     110
 *                                               ______________
 * WE    _______________________________________|              |__
 * 
 *                                              @
 * (RAM_USE_CLK)
 *                                       ______________
 * WE_mem ______________________________|              |__
 *              @               @               @               @
 * 
*/
module sram # (
                parameter SIZE = 16
//              parameter SIZE = 64
//              parameter SIZE = 128
//		parameter SIZE = 256
) (
                input CLK,
		input WEB,
                input [5:0] ADDRB,
                input [7:0] DIB,
                output [7:0] DOB
);
    reg  [7:0] mem [0:SIZE-1];
`ifdef RAM_USE_CLK
    reg [5:0] r_addr;
    always @(posedge CLK)
    if (SIZE == 16)
      begin
        if (WEB) mem[ADDRB[3:0]] <= DIB;
	r_addr <= { 2'b00, ADDRB[3:0] };
      end
    else
      begin
        if (WEB) mem[ADDRB] <= DIB;
	r_addr <= ADDRB;
      end
    assign DOB = mem[r_addr];
`else
    always @(posedge WEB)
    begin
        mem[ADDRB] <= DIB;
    end
    assign DOB = mem[ADDRB];
`endif

`ifdef RAM_INCLUDE_DAT
  initial
    begin
	if (SIZE == 64)
           $readmemh("mcpu_ram64.mem" , mem, 0 , SIZE-1);
	else
           $readmemh("mcpu_ram.mem" , mem, 0 , SIZE-1);
    end
`endif

endmodule

module rom # (
                parameter SIZE = 64
//              parameter SIZE = 128
//		parameter SIZE = 256
) (
                input CLK,
                input RD,
                input [6:0] ADDRA,
                output [11:0] DOA
);
    reg  [11:0] mem [0:SIZE-1];
`ifdef ROM_USE_CLK
    reg [6:0] r_addr;
    always @(negedge CLK)
    begin
	if (RD) r_addr <= ADDRA;
    end
    assign DOA = mem[r_addr];
`else
    assign DOA = mem[ADDRA];
`endif

  initial
    begin
	if (SIZE == 128)
	    $readmemh("mcpu3_code128.mem" , mem, 0 , SIZE-1);
	else
	    $readmemh("mcpu3_code.mem" , mem, 0 , SIZE-1);
    end

endmodule
