/*
 * sample driver for V-USB 
 *
 *  avr-gcc -mmcu=atmega88 -Os -c udrv-driver.c 
 *
 */

#include <avr/interrupt.h>
#include <avr/io.h>

static inline void  bit_set (uint8_t val , int8_t bit) {
	val |= (1<<bit);
}

static inline void  bit_clear (uint8_t val , int8_t bit) {
	val &= ~(1<<bit);
}

#if 0
static inline int8_t  bit_is_set (uint8_t val , int8_t bit) {
	return (val & (1<<bit)) != 0;
}

static inline int8_t  bit_is_clear (uint8_t val , int8_t bit) {
	return (val & (1<<bit)) == 0;
}

extern void spi_write(uint8_t);
extern uint8_t spi_write_read(uint8_t);
extern uint8_t spi_read();
extern int8_t spi_busy();

#else //  AVR depended functions

static inline int8_t spi_busy() {
	return !(SPSR & (1<<SPIF));
}

static inline void spi_write(uint8_t data) {
	while (spi_busy())
		;
	SPDR = data;
}

static inline uint8_t spi_write_read(uint8_t data) {
	SPDR = data;
	while (spi_busy())
		;
	return SPDR;
}

static inline uint8_t spi_read() {
	return spi_write_read(0x00);
}
#endif

extern volatile int8_t  usbRxLen;
extern uint8_t          usbInputBufOffset;
#define USB_BUFSIZE     11  /* original: PID, 8 bytes data, 2 bytes CRC */
extern uint8_t		usbRxBuf[USB_BUFSIZE+3];

extern uint8_t          usbCurrentTok; // last token ??
extern uint8_t          usbRxToken;
extern uint8_t          usbDeviceAddr;
extern uint8_t          usbNewDeviceAddr;

#if USB_CFG_CHECK_DATA_TOGGLING
extern uint8_t          usbCurrentDataToken;
#endif

/* token */
#define USBPID_OUT      0xe1 /* 1110 0001 */
#define USBPID_SOF      0xa5 /* 1010 0101 */
#define USBPID_IN       0x69 /* 0110 1001 */
#define USBPID_SETUP    0x2d /* 0010 1101 */
/* data */
#define USBPID_DATA0    0xc3 /* 1100 0011 */
#define USBPID_DATA1    0x4b /* 0100 1011 */
/* handshake */
#define USBPID_ACK      0xd2 /* 1101 0010 */
#define USBPID_NAK      0x5a /* 0101 1010 */
#define USBPID_STALL    0x1e /* 0001 1110 */

#define SENDRQ_NAK	(0x80 | (USBPID_NAK >> 1))
/*
usbRxLen : 
	 >0 : usbRxBuf is valid
	 0  : usbRxBuf is free
	 <0 : usbAllRequestsAreDisabled()

    usbdrv.h:#define usbDisableAllRequests()     usbRxLen = -1
    usbdrv.h:#define usbEnableAllRequests()      usbRxLen = 0
    usbdrv.h:#define usbAllRequestsAreDisabled() (usbRxLen < 0)

usbRxToken :
	USBPID_SETUP 0x2d ( 0010 1101 )
	       or
	endpoint number when USBPID_OUT

      if(usbRxToken < 0x10){ 
           usbFunctionWriteOut(data, len);
            return;
      }
      if(usbRxToken == (uchar)USBPID_SETUP){
	      :

*/
typedef struct usbTxStatus{
	volatile  int8_t   len;
	uint8_t   buffer[USB_BUFSIZE];
}usbTxStatus_t;

extern usbTxStatus_t   usbTxStatus1, usbTxStatus3;
#define usbTxLen1   usbTxStatus1.len
#define usbTxBuf1   usbTxStatus1.buffer
#define usbTxLen3   usbTxStatus3.len
#define usbTxBuf3   usbTxStatus3.buffer

extern volatile int8_t  usbTxLen;
extern uint8_t          usbTxBuf[USB_BUFSIZE];

/*

usbTxLen:

	buffer len (SYNC ... CRC-16)  : DATA packet only 
		or
	handshake PID : original ( only useful with Low-Speed (maxps == 8) )
	0x80 | (handshake PID >> 1) : udrv ( this can support maxps == 64 )

	need changing the followings :

		usbTxLen = USBPID_NAK;	     -> SENDRQ_NAK
		usbTxLen = USBPID_STALL;     -> SENDRQ_STALL
		if(usbTxLen & 0x10){         -> 0x80
		if(txStatus->len & 0x10){    -> 0x80
*/

#define USB_CFG_EP3_NUMBER 3
#define nDATA_RDY_PIN 	PINB
#define nDATA_RDY_BIT 	PB0
#define RS_PORT		PORTB
#define RS_BIT		PB1
#define nCS_PORT 	PORTB
#define nCS_BIT 	PB2

#define nIRQ_PIN	PIND
#define nIRQ_BIT	PD2

#define CTRL_READ_STATUS 0x00
#define CTRL_RECV_MODE	 0x83
#define CTRL_SEND_MODE	 0x80
#define CTRL_SEND_START	 0x82

static inline void data_write(int8_t count, uint8_t *datap) {
	bit_set(RS_PORT, RS_BIT);

	bit_clear(nCS_PORT, nCS_BIT);
	while (--count >= 0) {
		spi_write(*datap++);
	}
	while (spi_busy())
		;
	bit_set(nCS_PORT, nCS_BIT);
}

#define ST_CRC_ERROR	(1<<7)
#define ST_DATA_FULL	(1<<6)
#define ST_UNDER_HALF	(1<<5)
#define ST_DATA_RDY	(1<<4)
#define ST_BUSY		(1<<3)
#define ST_IRQ		(1<<2)
#define ST_RECV_MODE	(1<<1)
#define ST_START	(1<<0)

static inline uint8_t ctrl_write(uint8_t data) {
	uint8_t st;
	while (spi_busy())
		;
	bit_clear(RS_PORT, RS_BIT);
	bit_clear(nCS_PORT, nCS_BIT);
	while (spi_busy())
		;
	st = spi_write_read(data);
	bit_set(nCS_PORT, nCS_BIT);
	bit_set(RS_PORT, RS_BIT);
	return st;
}

// send ACK/NAK/STALL
static inline void send_handshake(uint8_t pid) {
	ctrl_write(CTRL_SEND_MODE);

	bit_clear(nCS_PORT, nCS_BIT);
	spi_write(pid);
	while (spi_busy())
		;
	bit_set(nCS_PORT, nCS_BIT);

	ctrl_write(CTRL_SEND_START);
	bit_set(EIFR, INTF0); // clear INT0 req
}

ISR(INT0_vect, ISR_BLOCK) {
	static uint8_t recv_offset = 0;
	static int8_t send_len = 0;
	static uint8_t *send_ptr;
	uint8_t i;
	uint16_t addr_ep;

#if 0
	if (send_len > 0) {
		bit_clear(nCS_PORT, nCS_BIT);
	        do {
			spi_write(*send_ptr++);
	        } while  ((--send_len > 0) && bit_is_clear(nIRQ_PIN, nIRQ_BIT));
		bit_set(EIFR, INTF0); // clear INT0 req
		goto re_send;
	}
#endif
	if (bit_is_clear(nDATA_RDY_PIN, nDATA_RDY_BIT)) {
re_recv:
	    do {
		i = spi_read();
		if (recv_offset == 0) usbCurrentTok = i;
		if (recv_offset < USB_BUFSIZE) {
			usbRxBuf[recv_offset++] = i;
		}
		if (usbCurrentTok & 0x10)
			goto break_eop; // handshake
		if ((usbCurrentTok & 0x21) && (recv_offset == 3))
			goto break_eop; // token
	    } while  (bit_is_clear(nDATA_RDY_PIN, nDATA_RDY_BIT));
	}
	i = ctrl_write(CTRL_READ_STATUS);
	if (i & ST_BUSY) return; // RECV in progress
	if (i & ST_DATA_RDY) goto re_recv;
	if (i & ST_CRC_ERROR) goto send_none;
	if (i & ST_DATA_FULL) {
		ctrl_write(CTRL_SEND_MODE);
		ctrl_write(CTRL_RECV_MODE);
		goto send_none;
	}
break_eop:
	addr_ep = *(uint16_t *)(usbRxBuf + 1);
	addr_ep <<= 1;
	i = (addr_ep >> 8) & 0xf;
	if ((usbCurrentTok == USBPID_SETUP) | (usbCurrentTok == USBPID_OUT)) {
		if ( recv_offset < 3 ) {
			if (usbRxLen) recv_offset = USB_BUFSIZE + 1;
			return; // RECV DATA0/1
		}
		if ((usbCurrentTok == USBPID_OUT)
		   && (recv_offset == USB_BUFSIZE + 1) ) { // buffer busy
			recv_offset = 0;
			goto send_nak;
		}
		if (recv_offset >= USB_BUFSIZE ) { // buffer overrun
			goto send_none;
		}
#if USB_CFG_CHECK_DATA_TOGGLING
		usbCurrentDataToken = usbRxBuf[3];
#endif
		if ((uint8_t)addr_ep != usbDeviceAddr)  goto send_nak;
		if (i) usbCurrentTok = i;
		usbRxToken = usbCurrentTok;
		// data top :  USB_BUFSIZE - usbInputBufOffset
		usbInputBufOffset = USB_BUFSIZE - 3;
		usbRxLen = recv_offset - 3;
		send_handshake(USBPID_ACK);
		goto send_none;
	} else if ( usbCurrentTok == USBPID_IN ) {
		if ((uint8_t)addr_ep != usbDeviceAddr)  goto send_nak;
		if (i == 0) {
			send_len = usbTxLen;
			send_ptr = usbTxBuf;
			usbTxLen = SENDRQ_NAK;
		} else if ( i == USB_CFG_EP3_NUMBER) {
			send_len = usbTxLen3;
			send_ptr = usbTxBuf3;
			usbTxLen3 = SENDRQ_NAK;
		} else {
			send_len = usbTxLen1;
			send_ptr = usbTxBuf1;
			usbTxLen1 = SENDRQ_NAK;
		}
		if (send_len < 0) {
			send_handshake(send_len << 1);
			send_len = 0;
			recv_offset = 0;
			return;
		}
		if (send_len > 3) send_len -= 3; // remove SYNC , CRC-16
		usbDeviceAddr = (usbNewDeviceAddr << 1);
		ctrl_write(CTRL_SEND_MODE);
		bit_clear(nCS_PORT, nCS_BIT);
	        do {
			spi_write(*send_ptr++);
	        } while  ((--send_len > 0) && bit_is_set(nIRQ_PIN, nIRQ_BIT));
		while (spi_busy())
			;
		bit_set(nCS_PORT, nCS_BIT);
		ctrl_write(CTRL_SEND_START);
		bit_set(EIFR, INTF0); // clear INT0 req
re_send:
		if (send_len > 0) {
			bit_clear(nCS_PORT, nCS_BIT);
			i = 20;
			if (send_len < i) i = send_len + 1;
			while (--i) {
				spi_write(*send_ptr++);
				--send_len;
			}
			while (spi_busy())
				;
		}
		bit_set(nCS_PORT, nCS_BIT);
	        if ((send_len == 0) || bit_is_clear(nIRQ_PIN, nIRQ_BIT))
			bit_set(EIFR, INTF0); // clear INT0 req
		return;
	}
	return;
send_nak:
	send_handshake(USBPID_NAK);
send_none:
	recv_offset = 0;
	if (bit_is_clear(nDATA_RDY_PIN, nDATA_RDY_BIT)) goto re_recv;
	return;
}
